#
# Virtual Lab driver for SMBH
#
# Copyright (C) 2010, Jack Whitham
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import vlab

def setPLL(vl, freq):
    """Program the SMBH clock chip.

    freq must be specified in MHz. The valid range is approximately
    0.01 to 250MHz. The clock signal is delivered to all FPGAs.
    The default frequency is 100MHz."""
    assert isinstance(vl, vlab.VlabChannel)
    return vl.hardwareSpecificCommand("pll", freq)

def masterReset(vl):
    """SMBH FPGA master reset.

    This clears the FPGA configurations, and resets the clock chip
    to its default setting of 100MHz. A master reset may be necessary
    before the FPGAs can be reprogrammed."""
    assert isinstance(vl, vlab.VlabChannel)
    return vl.hardwareSpecificCommand("reset")

def softwareReset(vl, fpga_num, reset_state):
    """Send software reset signal to a specific FPGA.

    This controls the level of the RESET input pin, delivered to each FPGA.
    It is intended to be used to reset a CPU, but it can be used for any
    purpose, including as a GPIO. reset_state may be 0 or 1. fpga_num
    must be 0, 1, 2, or 3."""
    assert isinstance(vl, vlab.VlabChannel)
    return vl.hardwareSpecificCommand("setgpio", fpga_num, reset_state)
        
