/*
 * Scratchpad MMU - JPEG case study main module
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#undef NDEBUG
#include <stdio.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <assert.h>

/* Xilinx */
#include "xparameters.h"
#include "xuartns550_l.h"
#include "xutil.h"
#include "mb_interface.h"

/* libjpeg */
#include "jinclude.h"
#include "jpeglib.h"
#include "jerror.h"

/* SMMU */
#include "drvif.h"
#include "spad.h"

#define NUM_ROWS_AT_ONCE 4

/*#undef print
#define print(x)*/


extern const char input_large_data [] ;
extern int input_large_len ;
int         max_at_once ;


/* Expanded data source object for stdio input */

typedef struct {
  struct jpeg_source_mgr pub;	/* public fields */

  boolean start_of_file;	/* have we gotten any data yet? */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

METHODDEF(void)
init_source (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  /* We reset the empty-input-file flag for each image,
   * but we don't clear the input buffer.
   * This is correct behavior for reading a series of images from one source.
   */
  src->start_of_file = TRUE;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

METHODDEF(boolean)
fill_input_buffer (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  src->pub.next_input_byte = (const unsigned char *) & input_large_data [ 0 ] ;
  src->pub.bytes_in_buffer = input_large_len ;
  src->start_of_file = FALSE;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

METHODDEF(void)
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  /*my_src_ptr src = (my_src_ptr) cinfo->src;*/

  assert ( 0 ) ;
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

METHODDEF(void)
term_source (j_decompress_ptr cinfo)
{
  /* no work necessary here */
}


/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

static my_src_ptr already_done = NULL ;
void memory_source (j_decompress_ptr cinfo)
{
  my_src_ptr src;

  /* The source object and input buffer are made permanent so that a series
   * of JPEG images can be read from the same file by calling jpeg_stdio_src
   * only before the first one.  (If we discarded the buffer at the end of
   * one image, we'd likely lose the start of the next one.)
   * This makes it unsafe to use this manager and a different source
   * manager serially with the same JPEG object.  Caveat programmer.
   */
  if (cinfo->src == NULL) {	/* first time for this JPEG object? */
    /*if ( already_done != NULL )
    {
      print ( "Allocate again.\r\n" ) ;
      src = already_done ;
      cinfo->src = src ;
    } else {
      print ( "Allocate first time.\r\n" ) ;*/
      cinfo->src = (struct jpeg_source_mgr *)
        (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				    SIZEOF(my_source_mgr));
      already_done = src = (my_src_ptr) cinfo->src;
    /*}*/
  }

  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
  src->pub.term_source = term_source;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
  src->pub.next_input_byte = NULL; /* until buffer loaded */
}


struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
};

typedef struct my_error_mgr * my_error_ptr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
    /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
    /*my_error_ptr myerr = (my_error_ptr) cinfo->err;*/

    /* Always display the message. */
    /* We could postpone this until after returning, if we chose. */
    (*cinfo->err->output_message) (cinfo);

    exit ( 1 ) ;
}

static struct my_error_mgr jerr;
/* More stuff */
static JSAMPARRAY buffer;		/* Output row buffer */
static int row_stride;		/* physical row width in output buffer */
static struct jpeg_decompress_struct cinfo ;
static unsigned sum = 0 ;

void Initialise ( void )
{   
    /* Step 1: allocate and initialize JPEG decompression object */
    /* We set up the normal JPEG error routines, then override error_exit. */
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;

    /* Now we can initialize the JPEG decompression object. */
    jpeg_create_decompress(&cinfo);


    /* Step 2: specify data source (eg, a file) */
    memory_source(&cinfo);     /* XXX */

    /* Step 3: read file parameters with jpeg_read_header() */

    (void) jpeg_read_header(&cinfo, TRUE);
    /* We can ignore the return value from jpeg_read_header since
    *   (a) suspension is not possible with the stdio data source, and
    *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
    * See libjpeg.doc for more info.
    */

    /* Step 4: set parameters for decompression */
    cinfo.dct_method = JDCT_IFAST ; /* software */
    cinfo.quantize_colors = FALSE ;
    cinfo.out_color_space = JCS_RGB;
}

void Benchmark ( void )
{
    int *   check1 ;
    int *   check2 ;
    int *   check3 ;
    int     i ;

    sum = 0 ;

    /* Step 5: Start decompressor */

    (void) jpeg_start_decompress(&cinfo);
    /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */

    /* We may need to do some setup of our own at this point before reading
    * the data.  After jpeg_start_decompress() we have the correct scaled
    * output image dimensions available, as well as the output colormap
    * if we asked for color quantization.
    * In this example, we need to make an output work buffer of the right size.
    */ 
    /* JSAMPLEs per row in output buffer */
    row_stride = cinfo.output_width * cinfo.output_components;
    /* Make a one-row-high sample array that will go away when done with image */
    buffer = (*cinfo.mem->alloc_sarray)
        ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, NUM_ROWS_AT_ONCE);


    /* Step 6: while (scan lines remain to be read) */
    /*           jpeg_read_scanlines(...); */

    /* Here we use the library's state variable cinfo.output_scanline as the
    * loop counter, so that we don't have to keep track ourselves.
    */
    while (cinfo.output_scanline < cinfo.output_height) {
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
     * Here the array is only one element long, but you could ask for
     * more than one scanline at a time if that's more convenient.
     */
        int done = jpeg_read_scanlines(&cinfo, buffer, NUM_ROWS_AT_ONCE );

        for ( i = 0 ; i < done ; i ++ )
        {
            check1 = (int *) & buffer [ i ][ 0 ] ;
            check2 = (int *) & buffer [ i ][ ( row_stride / 2 ) & ~3 ] ;
            check3 = (int *) & buffer [ i ][ ( row_stride - 4 ) & ~3 ] ;
            sum = sum + (* check1) + (* check2) + (* check3) ;
        }
        if ( done > max_at_once )
        {
            max_at_once = done ;
        }
    }

    /* Step 7: Finish decompression */

    (void) jpeg_finish_decompress(&cinfo);
    /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */
}
void printhex ( unsigned x )
{
    unsigned i , y ;
    char out [ 2 ] ;

    out [ 1 ] = '\0' ;
    for ( i = 0 ; i < 32 ; i += 4 )
    {
        y = ( x << i ) >> 28 ;
        if ( y < 10 )
        {
            out [ 0 ] = y + '0' ;
        } else {
            out [ 0 ] = y + 'a' - 10 ;
        }
        print ( out ) ;
    }
} 

void Execute ( void )
{
    sum = 0 ;
    max_at_once = 0 ;

    print ( "djpeg running: " ) ;
    Initialise () ;
    Benchmark () ;
    jpeg_destroy_decompress(&cinfo);

    if (( jerr . pub . num_warnings == 0 )
    && ( sum == 0xe7341acc ))
    {
        print ( "correct!\r\n" ) ;
        print ( "max at once " ) ;
        printhex ( max_at_once ) ;
        print ( "\r\n" ) ;
    } else {
        print ( "failure!\r\n" ) ;
        print ( "checksum " ) ;
        printhex ( jerr . pub . num_warnings ) ;
        printhex ( sum ) ;
        print ( "\r\n" ) ;
    }
}

void Table_Out ( const char * name , unsigned index )
{
    print ( "\r\n[" ) ;
    print ( (char *) name ) ;
    print ( "] global_counter = " ) ;
    printhex ( _ar_id [ index ] . global_counter ) ;
    print ( "\r\nmatch_count = " ) ;
    printhex ( _ar_id [ index ] . match_count ) ;
    print ( "\r\nmiss_count = " ) ;
    printhex ( _ar_id [ index ] . miss_count ) ;
    print ( "\r\ncall_count = " ) ;
    printhex ( _ar_id [ index ] . call_count ) ;
}

int main ( void )
{
    Activity_Record start_time ;
    Activity_Record total_time ;

    XUartNs550_SetBaud(XPAR_RS232_UART_1_BASEADDR, 
            XPAR_XUARTNS550_CLOCK_HZ, 115200);
    XUartNs550_mSetLineControlReg(XPAR_RS232_UART_1_BASEADDR, 
            XUN_LCR_8_DATA_BITS);
    print ( "spad " ) ;
    Zero_Spad () ;

    print ( "smmu " ) ;
    HW_Init () ;
    read_only_remaining = NUM_RO_ENTRIES;

    print ( "cache " ) ;

    //print ( "djpeg relocating\r\n" ) ;
    //Relocate () ;
#if XPAR_MICROBLAZE_0_USE_ICACHE
    microblaze_init_icache_range(0, XPAR_MICROBLAZE_0_CACHE_BYTE_SIZE);
    microblaze_enable_icache();
#endif
    print ( "ok\r\n" ) ;

    while ( 1 )
    {
        memset ( & total_time , 0 , sizeof ( Activity_Record ) ) ;
        memset ( & start_time , 0 , sizeof ( Activity_Record ) ) ;
        memset ( _ar_id , 0 , sizeof ( _ar_id ) ) ;
        print ( "\r\n\r\n" ) ;
        printhex ( activate_smmu ) ;
        print ( "\r\n" ) ;
        Capture_Activity_Record ( & start_time , & start_time ) ;
        Execute () ;
        Capture_Activity_Record ( & total_time , & start_time ) ;
        print ( "[ENV] " ) ;
        if ( activate_smmu & ACT_DCT ) print ( "DCT " ) ;
        if ( activate_smmu & ACT_DCTOUT ) print ( "DCTOUT " ) ;
        if ( activate_smmu & ACT_YCC ) print ( "YCC " ) ;
        if ( activate_smmu & ACT_DEC ) print ( "DEC " ) ;
        if ( activate_smmu & ACT_DEC1 ) print ( "DEC1 " ) ;
        Table_Out ( "DCT" , TIME_DCT ) ;
        Table_Out ( "CC" , TIME_CC ) ;
        Table_Out ( "DCT1" , TIME_DCT1 ) ;
        Table_Out ( "DEC" , TIME_DEC ) ;

        print ( "\r\ntotal time = " ) ;
        printhex ( total_time . global_counter ) ;
        print ( "\r\n" ) ;

        asm volatile ( ".long 0x80000001\n" ) ;
        print ( "\r\nThis is not a simulator! Going again.\r\n" ) ;
        activate_smmu ^= 0xff ;
    }
    return 0 ;
}

void __assert (const char * one, int two , const char * three)
{
    print("\r\nBang! Assertion failed\r\n");
    print( (char *) one);
    print(" ");
    printhex(two);
    print(" ");
    print( (char *) three);
    print("\r\n");
    asm volatile ( ".long 0x80000001\n" ) ;
    exit ( 1 ) ;
}


