--
-- Scratchpad MMU - PLB and LMB interfaces for SMMU
-- Copyright (C) Jack Whitham 2009
-- http://www.jwhitham.org.uk/c/smmu.html
--
-- This library is free software; you can redistribute it and/or
-- modify it under the terms of the GNU Lesser General Public
-- License as published by the Free Software Foundation
-- (version 2.1 of the License only).
-- 
-- This library is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-- Lesser General Public License for more details.
-- 
-- You should have received a copy of the GNU Lesser General Public
-- License along with this library; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
-- 
-- Template generated Sun Mar 08 15:13:58 2009 by Create and 
-- Import Peripheral Wizard

-- DO NOT EDIT BELOW THIS LINE --------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library proc_common_v2_00_a;
use proc_common_v2_00_a.proc_common_pkg.all;
use proc_common_v2_00_a.srl_fifo_f;


-- DO NOT EDIT ABOVE THIS LINE --------------------


--USER libraries added here

------------------------------------------------------------------------------
-- Entity section
------------------------------------------------------------------------------
-- Definition of Generics:
--   C_SLV_DWIDTH                 -- Slave interface data bus width
--   C_MST_AWIDTH                 -- Master interface address bus width
--   C_MST_DWIDTH                 -- Master interface data bus width
--   C_NUM_REG                    -- Number of software accessible registers
--
-- Definition of Ports:
--   Bus2IP_Clk                   -- Bus to IP clock
--   Bus2IP_Reset                 -- Bus to IP reset
--   Bus2IP_Data                  -- Bus to IP data bus
--   Bus2IP_BE                    -- Bus to IP byte enables
--   Bus2IP_RdCE                  -- Bus to IP read chip enable
--   Bus2IP_WrCE                  -- Bus to IP write chip enable
--   IP2Bus_Data                  -- IP to Bus data bus
--   IP2Bus_RdAck                 -- IP to Bus read transfer acknowledgement
--   IP2Bus_WrAck                 -- IP to Bus write transfer acknowledgement
--   IP2Bus_Error                 -- IP to Bus error response
--   IP2Bus_MstRd_Req             -- IP to Bus master read request
--   IP2Bus_MstWr_Req             -- IP to Bus master write request
--   IP2Bus_Mst_Addr              -- IP to Bus master address bus
--   IP2Bus_Mst_BE                -- IP to Bus master byte enables
--   IP2Bus_Mst_Length            -- IP to Bus master transfer length
--   IP2Bus_Mst_Type              -- IP to Bus master transfer type
--   IP2Bus_Mst_Lock              -- IP to Bus master lock
--   IP2Bus_Mst_Reset             -- IP to Bus master reset
--   Bus2IP_Mst_CmdAck            -- Bus to IP master command acknowledgement
--   Bus2IP_Mst_Cmplt             -- Bus to IP master transfer completion
--   Bus2IP_Mst_Error             -- Bus to IP master error response
--   Bus2IP_Mst_Rearbitrate       -- Bus to IP master re-arbitrate
--   Bus2IP_Mst_Cmd_Timeout       -- Bus to IP master command timeout
--   Bus2IP_MstRd_d               -- Bus to IP master read data bus
--   Bus2IP_MstRd_rem             -- Bus to IP master read remainder
--   Bus2IP_MstRd_sof_n           -- Bus to IP master read start of frame
--   Bus2IP_MstRd_eof_n           -- Bus to IP master read end of frame
--   Bus2IP_MstRd_src_rdy_n       -- Bus to IP master read source ready
--   Bus2IP_MstRd_src_dsc_n       -- Bus to IP master read source discontinue
--   IP2Bus_MstRd_dst_rdy_n       -- IP to Bus master read destination ready
--   IP2Bus_MstRd_dst_dsc_n       -- IP to Bus master read destination discontinue
--   IP2Bus_MstWr_d               -- IP to Bus master write data bus
--   IP2Bus_MstWr_rem             -- IP to Bus master write remainder
--   IP2Bus_MstWr_sof_n           -- IP to Bus master write start of frame
--   IP2Bus_MstWr_eof_n           -- IP to Bus master write end of frame
--   IP2Bus_MstWr_src_rdy_n       -- IP to Bus master write source ready
--   IP2Bus_MstWr_src_dsc_n       -- IP to Bus master write source discontinue
--   Bus2IP_MstWr_dst_rdy_n       -- Bus to IP master write destination ready
--   Bus2IP_MstWr_dst_dsc_n       -- Bus to IP master write destination discontinue
------------------------------------------------------------------------------

entity user_logic is
  generic
  (
    -- Xilinx things
    C_LMB_HIGHADDR   : std_logic_vector(0 to 31) := X"00000000";
    C_LMB_BASEADDR   : std_logic_vector(0 to 31) := X"FFFFFFFF";
    C_LMB_AWIDTH : integer                   := 32;
    C_LMB_DWIDTH : integer                   := 32;
    C_SLV_DWIDTH                   : integer              := 32;
    C_MST_AWIDTH                   : integer              := 32;
    C_MST_DWIDTH                   : integer              := 32;
    C_NUM_REG                      : integer              := 8;

    -- SMMU configuration
    ENABLE_EXTRA_SCRATCHPAD         : Boolean;  -- true
    LOG2_NUM_CACHE_LINES_PER_PAGE   : Natural;
    LOG2_NUM_CACHE_WORDS_PER_LINE   : Natural;
    LOG2_NUM_ENTRIES                : Natural;  -- 4
    NUM_ACTIVE_BITS                 : Natural;  -- 24
    NUM_READ_ONLY_ENTRIES           : Natural;   -- 3
    INITIATE_BOOST                  : Natural;   -- 1
    C_CONTROL_MASK    : std_logic_vector ( 0 to 31 );
    C_USE_MASK        : std_logic_vector ( 0 to 31 );
    C_SP_MASK         : std_logic_vector ( 0 to 31 )
  );
  port
  (
    -- ADD USER PORTS BELOW THIS LINE ------------------
    --USER ports added here
    LMB_Clk : in std_logic := '0';
    LMB_Rst : in std_logic := '0';

    -- Slave Local Memory Bus
    LMB_ABus        : in  std_logic_vector(0 to C_LMB_AWIDTH-1);
    LMB_WriteDBus   : in  std_logic_vector(0 to C_LMB_DWIDTH-1);
    LMB_AddrStrobe  : in  std_logic;
    LMB_ReadStrobe  : in  std_logic;
    LMB_WriteStrobe : in  std_logic;
    LMB_BE          : in  std_logic_vector(0 to (C_LMB_DWIDTH/8 - 1));
    Sl_DBus         : out std_logic_vector(0 to C_LMB_DWIDTH-1);
    Sl_Ready        : out std_logic;
    -- ADD USER PORTS ABOVE THIS LINE ------------------

    -- DO NOT EDIT BELOW THIS LINE ---------------------
    -- Bus protocol ports, do not add to or delete
    Bus2IP_Clk                     : in  std_logic;
    Bus2IP_Reset                   : in  std_logic;
    Bus2IP_Data                    : in  std_logic_vector(0 to C_SLV_DWIDTH-1);
    Bus2IP_BE                      : in  std_logic_vector(0 to C_SLV_DWIDTH/8-1);
    Bus2IP_RdCE                    : in  std_logic_vector(0 to C_NUM_REG-1);
    Bus2IP_WrCE                    : in  std_logic_vector(0 to C_NUM_REG-1);
    IP2Bus_Data                    : out std_logic_vector(0 to C_SLV_DWIDTH-1);
    IP2Bus_RdAck                   : out std_logic;
    IP2Bus_WrAck                   : out std_logic;
    IP2Bus_Error                   : out std_logic;
    IP2Bus_MstRd_Req               : out std_logic;
    IP2Bus_MstWr_Req               : out std_logic;
    IP2Bus_Mst_Addr                : out std_logic_vector(0 to C_MST_AWIDTH-1);
    IP2Bus_Mst_BE                  : out std_logic_vector(0 to C_MST_DWIDTH/8-1);
    IP2Bus_Mst_Length              : out std_logic_vector(0 to 11);
    IP2Bus_Mst_Type                : out std_logic;
    IP2Bus_Mst_Lock                : out std_logic;
    IP2Bus_Mst_Reset               : out std_logic;
    Bus2IP_Mst_CmdAck              : in  std_logic;
    Bus2IP_Mst_Cmplt               : in  std_logic;
    Bus2IP_Mst_Error               : in  std_logic;
    Bus2IP_Mst_Rearbitrate         : in  std_logic;
    Bus2IP_Mst_Cmd_Timeout         : in  std_logic;
    Bus2IP_MstRd_d                 : in  std_logic_vector(0 to C_MST_DWIDTH-1);
    Bus2IP_MstRd_rem               : in  std_logic_vector(0 to C_MST_DWIDTH/8-1);
    Bus2IP_MstRd_sof_n             : in  std_logic;
    Bus2IP_MstRd_eof_n             : in  std_logic;
    Bus2IP_MstRd_src_rdy_n         : in  std_logic;
    Bus2IP_MstRd_src_dsc_n         : in  std_logic;
    IP2Bus_MstRd_dst_rdy_n         : out std_logic;
    IP2Bus_MstRd_dst_dsc_n         : out std_logic;
    IP2Bus_MstWr_d                 : out std_logic_vector(0 to C_MST_DWIDTH-1);
    IP2Bus_MstWr_rem               : out std_logic_vector(0 to C_MST_DWIDTH/8-1);
    IP2Bus_MstWr_sof_n             : out std_logic;
    IP2Bus_MstWr_eof_n             : out std_logic;
    IP2Bus_MstWr_src_rdy_n         : out std_logic;
    IP2Bus_MstWr_src_dsc_n         : out std_logic;
    Bus2IP_MstWr_dst_rdy_n         : in  std_logic;
    Bus2IP_MstWr_dst_dsc_n         : in  std_logic
    -- DO NOT EDIT ABOVE THIS LINE ---------------------
  );

  attribute SIGIS : string;
  attribute SIGIS of Bus2IP_Clk    : signal is "CLK";
  attribute SIGIS of Bus2IP_Reset  : signal is "RST";
  attribute SIGIS of IP2Bus_Mst_Reset: signal is "RST";

end entity user_logic;

------------------------------------------------------------------------------
-- Architecture section
------------------------------------------------------------------------------

architecture IMP of user_logic is

  -- These "options" cannot be easily changed.
  constant CPU_LEFT         : Natural := 31 ;
  constant SP_LEFT          : Natural := 10 ;
  constant DATA_LEFT        : Natural := 31 ;

  -- These are derived:
  constant NUM_ENTRIES      : Natural := 2 ** LOG2_NUM_ENTRIES ;
  constant INACTIVE         : Natural := 31 - NUM_ACTIVE_BITS ;
  constant CONTROL_MASK     : std_logic_vector ( 
        CPU_LEFT downto NUM_ACTIVE_BITS  ) := C_CONTROL_MASK ( 0 to INACTIVE );
  constant USE_MASK         : std_logic_vector ( 
        CPU_LEFT downto NUM_ACTIVE_BITS  ) := C_USE_MASK ( 0 to INACTIVE );
  constant SP_MASK          : std_logic_vector ( 
        CPU_LEFT downto NUM_ACTIVE_BITS  ) := C_SP_MASK ( 0 to INACTIVE );

  --USER signal declarations added here, as needed for user logic
  signal cpu_address        : std_logic_vector(CPU_LEFT downto 0); -- in
  signal cpu_data_in        : std_logic_vector(DATA_LEFT downto 0); -- in
  signal cpu_data_out       : std_logic_vector(DATA_LEFT downto 0); -- out
  signal cpu_copy_loc       : std_logic_vector(SP_LEFT downto 0); -- in
  signal cpu_size           : std_logic_vector(SP_LEFT downto 0); -- in
  signal cpu_write          : std_logic; -- in
  signal cpu_read           : std_logic; -- in
  signal cpu_open           : std_logic; -- in
  signal cpu_close          : std_logic; -- in
  signal cpu_cache_set_perm : std_logic; -- in
  signal cpu_cache_invalidate: std_logic;-- in
  signal cpu_ack            : std_logic; -- out
  signal cpu_ben            : std_logic_vector(3 downto 0);
  signal cache_enable       : std_logic;

  signal configuration_register 
                            : std_logic_vector(DATA_LEFT downto 0);
  signal configuration_register_2 
                            : std_logic_vector(DATA_LEFT downto 0); 

  signal sp_cpu_data_out    : std_logic_vector(DATA_LEFT downto 0); -- out

  subtype Bus_CPU_Data is std_logic_vector ( DATA_LEFT downto 0 ) ;

  signal global_counter     : Bus_CPU_Data ;
  signal status_reg         : Bus_CPU_Data ;
  signal cache_ctrl_mode    : std_logic;
  signal ctrl_mode          : std_logic;
  signal ctrl_select        : std_logic;
  signal ctrl_ack           : std_logic;
  signal ctrl_sp_ack        : std_logic;
  signal ctrl_reg           : std_logic_vector ( 2 downto 0 ) ;
  signal result_counter     : Bus_CPU_Data ;
  signal result_timer       : Bus_CPU_Data ;
  signal match_counter      : Bus_CPU_Data ;
  signal miss_counter       : Bus_CPU_Data ;
  signal d_cpu_ack          : std_logic; 
  signal enable_counters    : std_logic;
  signal select_read_only   : std_logic;
  signal d_enable_counters  : std_logic;
  signal is_read_write      : std_logic;
  signal d_is_read_write    : std_logic;
  signal lmb_we             : std_logic_vector ( 0 to 3 ) ;
  signal lmb_select         : std_logic ;
  signal sp_select          : std_logic ;
  signal sp_ack             : std_logic ;
  signal LMB_ABus_int       : std_logic_vector ( CPU_LEFT downto 0 ) ;


  ------------------------------------------
  -- Signals for user logic slave model s/w accessible register example
  ------------------------------------------
  signal slv_reg_write_sel              : std_logic_vector(0 to 3);
  signal slv_reg_read_sel               : std_logic_vector(0 to 3);
  signal slv_ip2bus_data                : std_logic_vector(0 to C_SLV_DWIDTH-1);
  signal slv_read_ack                   : std_logic;
  signal slv_write_ack                  : std_logic;

  ------------------------------------------
  -- Signals for user logic master model example
  ------------------------------------------
  -- signals for master model control/status registers write/read
  signal mst_ip2bus_data                : std_logic_vector(0 to C_SLV_DWIDTH-1);
  signal mst_reg_write_req              : std_logic;
  signal mst_reg_read_req               : std_logic;
  signal mst_reg_write_sel              : std_logic_vector(0 to 3);
  signal mst_reg_read_sel               : std_logic_vector(0 to 3);
  signal mst_write_ack                  : std_logic;
  signal mst_read_ack                   : std_logic;
  -- signals for master model control/status registers
  type BYTE_REG_TYPE is array(0 to 15) of std_logic_vector(0 to 7);
  signal mst_reg                        : BYTE_REG_TYPE;
  signal mst_byte_we                    : std_logic_vector(0 to 15);
  signal mst_cntl_rd_req                : std_logic;
  signal mst_cntl_wr_req                : std_logic;
  signal mst_cntl_bus_lock              : std_logic;
  signal mst_cntl_burst                 : std_logic;
  signal mst_ip2bus_addr                : std_logic_vector(0 to C_MST_AWIDTH-1);
  signal mst_xfer_length                : std_logic_vector(0 to 11);
  signal mst_ip2bus_be                  : std_logic_vector(0 to 15);
  signal mst_go                         : std_logic;
  -- signals for master model command interface state machine
  type CMD_CNTL_SM_TYPE is (CMD_IDLE, CMD_RUN, CMD_WAIT_FOR_DATA, CMD_DONE);
  signal mst_cmd_sm_state               : CMD_CNTL_SM_TYPE;
  signal mst_cmd_sm_set_done            : std_logic;
  signal mst_cmd_sm_set_error           : std_logic;
  signal mst_cmd_sm_set_timeout         : std_logic;
  signal mst_cmd_sm_busy                : std_logic;
  signal mst_cmd_sm_clr_go              : std_logic;
  signal mst_cmd_sm_rd_req              : std_logic;
  signal mst_cmd_sm_wr_req              : std_logic;
  signal mst_cmd_sm_reset               : std_logic;
  signal mst_cmd_sm_bus_lock            : std_logic;
  signal mst_cmd_sm_ip2bus_addr         : std_logic_vector(0 to C_MST_AWIDTH-1);
  signal mst_cmd_sm_ip2bus_be           : std_logic_vector(0 to C_MST_DWIDTH/8-1);
  signal mst_cmd_sm_xfer_type           : std_logic;
  signal mst_cmd_sm_xfer_length         : std_logic_vector(0 to 11);
  signal mst_cmd_sm_start_rd_llink      : std_logic;
  signal mst_cmd_sm_start_wr_llink      : std_logic;
  -- signals for master model read locallink interface state machine
  type RD_LLINK_SM_TYPE is (LLRD_IDLE, LLRD_GO);
  signal mst_llrd_sm_state              : RD_LLINK_SM_TYPE;
  signal mst_llrd_sm_dst_rdy            : std_logic;
  -- signals for master model write locallink interface state machine
  type WR_LLINK_SM_TYPE is (LLWR_IDLE, LLWR_SNGL_INIT, LLWR_SNGL, LLWR_BRST_INIT, LLWR_BRST, LLWR_BRST_LAST_BEAT);
  signal mst_llwr_sm_state              : WR_LLINK_SM_TYPE;
  signal mst_llwr_sm_src_rdy            : std_logic;
  signal mst_llwr_sm_sof                : std_logic;
  signal mst_llwr_sm_eof                : std_logic;
  signal mst_llwr_byte_cnt              : integer range 0 to 4095;
  signal mst_fifo_valid_write_xfer      : std_logic;
  signal mst_fifo_valid_read_xfer       : std_logic;

  signal dbg_code                : std_logic_vector ( 7 downto 0 );

begin

  --USER logic implementation added here

  ------------------------------------------
  -- Example code to read/write user logic slave model s/w accessible registers
  --
  -- Note:
  -- The example code presented here is to show you one way of reading/writing
  -- software accessible registers implemented in the user logic slave model.
  -- Each bit of the Bus2IP_WrCE/Bus2IP_RdCE signals is configured to correspond
  -- to one software accessible register by the top level template. For example,
  -- if you have four 32 bit software accessible registers in the user logic,
  -- you are basically operating on the following memory mapped registers:
  --
  --    Bus2IP_WrCE/Bus2IP_RdCE   Memory Mapped Register
  --                     "1000"   C_BASEADDR + 0x0
  --                     "0100"   C_BASEADDR + 0x4
  --                     "0010"   C_BASEADDR + 0x8
  --                     "0001"   C_BASEADDR + 0xC
  --
  ------------------------------------------
  slv_write_ack     <= Bus2IP_WrCE(0) or Bus2IP_WrCE(1) or Bus2IP_WrCE(2) or Bus2IP_WrCE(3);
  slv_read_ack      <= Bus2IP_RdCE(0) or Bus2IP_RdCE(1) or Bus2IP_RdCE(2) or Bus2IP_RdCE(3);
  slv_ip2bus_data <= ( others => '0' ) ;

  ------------------------------------------
  mst_reg_write_req <= Bus2IP_WrCE(4) or Bus2IP_WrCE(5) or Bus2IP_WrCE(6) or Bus2IP_WrCE(7);
  mst_reg_read_req  <= Bus2IP_RdCE(4) or Bus2IP_RdCE(5) or Bus2IP_RdCE(6) or Bus2IP_RdCE(7);
  mst_reg_write_sel <= Bus2IP_WrCE(4 to 7);
  mst_reg_read_sel  <= Bus2IP_RdCE(4 to 7);
  mst_write_ack     <= mst_reg_write_req;
  mst_read_ack      <= mst_reg_read_req;


  -- user logic master command interface assignments
  IP2Bus_MstRd_Req  <= mst_cmd_sm_rd_req;
  IP2Bus_MstWr_Req  <= mst_cmd_sm_wr_req;
  IP2Bus_Mst_Addr   <= mst_cmd_sm_ip2bus_addr;
  IP2Bus_Mst_BE     <= mst_cmd_sm_ip2bus_be;
  IP2Bus_Mst_Type   <= mst_cmd_sm_xfer_type;
  IP2Bus_Mst_Length <= mst_cmd_sm_xfer_length;
  IP2Bus_Mst_Lock   <= mst_cmd_sm_bus_lock;
  IP2Bus_Mst_Reset  <= mst_cmd_sm_reset;

  --implement master command interface state machine
  MASTER_CMD_SM_PROC : process( Bus2IP_Clk ) is
  begin

    if ( Bus2IP_Clk'event and Bus2IP_Clk = '1' ) then
      if ( Bus2IP_Reset = '1' ) then

        -- reset condition
        mst_cmd_sm_state          <= CMD_IDLE;
        mst_cmd_sm_clr_go         <= '0';
        mst_cmd_sm_rd_req         <= '0';
        mst_cmd_sm_wr_req         <= '0';
        mst_cmd_sm_bus_lock       <= '0';
        mst_cmd_sm_reset          <= '0';
        mst_cmd_sm_ip2bus_addr    <= (others => '0');
        mst_cmd_sm_ip2bus_be      <= (others => '0');
        mst_cmd_sm_xfer_type      <= '0';
        mst_cmd_sm_xfer_length    <= (others => '0');
        mst_cmd_sm_set_done       <= '0';
        mst_cmd_sm_set_error      <= '0';
        mst_cmd_sm_set_timeout    <= '0';
        mst_cmd_sm_busy           <= '0';
        mst_cmd_sm_start_rd_llink <= '0';
        mst_cmd_sm_start_wr_llink <= '0';

      else

        -- default condition
        mst_cmd_sm_clr_go         <= '0';
        mst_cmd_sm_rd_req         <= '0';
        mst_cmd_sm_wr_req         <= '0';
        mst_cmd_sm_bus_lock       <= '0';
        mst_cmd_sm_reset          <= '0';
        mst_cmd_sm_ip2bus_addr    <= (others => '0');
        mst_cmd_sm_ip2bus_be      <= (others => '0');
        mst_cmd_sm_xfer_type      <= '0';
        mst_cmd_sm_xfer_length    <= (others => '0');
        mst_cmd_sm_set_done       <= '0';
        mst_cmd_sm_set_error      <= '0';
        mst_cmd_sm_set_timeout    <= '0';
        mst_cmd_sm_busy           <= '1';
        mst_cmd_sm_start_rd_llink <= '0';
        mst_cmd_sm_start_wr_llink <= '0';

        -- state transition
        case mst_cmd_sm_state is

          when CMD_IDLE =>
            if ( mst_go = '1' ) then
              mst_cmd_sm_state  <= CMD_RUN;
              mst_cmd_sm_clr_go <= '1';
              if ( mst_cntl_rd_req = '1' ) then
                mst_cmd_sm_start_rd_llink <= '1';
              elsif ( mst_cntl_wr_req = '1' ) then
                mst_cmd_sm_start_wr_llink <= '1';
              end if;
            else
              mst_cmd_sm_state  <= CMD_IDLE;
              mst_cmd_sm_busy   <= '0';
            end if;

          when CMD_RUN =>
            if ( Bus2IP_Mst_CmdAck = '1' and Bus2IP_Mst_Cmplt = '0' ) then
              mst_cmd_sm_state <= CMD_WAIT_FOR_DATA;
            elsif ( Bus2IP_Mst_Cmplt = '1' ) then
              mst_cmd_sm_state <= CMD_DONE;
              if ( Bus2IP_Mst_Cmd_Timeout = '1' ) then
                -- PLB address phase timeout
                mst_cmd_sm_set_error   <= '1';
                mst_cmd_sm_set_timeout <= '1';
              elsif ( Bus2IP_Mst_Error = '1' ) then
                -- PLB data transfer error
                mst_cmd_sm_set_error   <= '1';
              end if;
            else
              mst_cmd_sm_state       <= CMD_RUN;
              mst_cmd_sm_rd_req      <= mst_cntl_rd_req;
              mst_cmd_sm_wr_req      <= mst_cntl_wr_req;
              mst_cmd_sm_ip2bus_addr <= mst_ip2bus_addr;
              mst_cmd_sm_ip2bus_be   <= mst_ip2bus_be(16-C_MST_DWIDTH/8 to 15);
              mst_cmd_sm_xfer_type   <= mst_cntl_burst;
              mst_cmd_sm_xfer_length <= mst_xfer_length;
              mst_cmd_sm_bus_lock    <= mst_cntl_bus_lock;
            end if;

          when CMD_WAIT_FOR_DATA =>
            if ( Bus2IP_Mst_Cmplt = '1' ) then
              mst_cmd_sm_state <= CMD_DONE;
            else
              mst_cmd_sm_state <= CMD_WAIT_FOR_DATA;
            end if;

          when CMD_DONE =>
            mst_cmd_sm_state    <= CMD_IDLE;
            mst_cmd_sm_set_done <= '1';
            mst_cmd_sm_busy     <= '0';

          when others =>
            mst_cmd_sm_state    <= CMD_IDLE;
            mst_cmd_sm_busy     <= '0';

        end case;

      end if;
    end if;

  end process MASTER_CMD_SM_PROC;

  IP2Bus_MstRd_dst_rdy_n <= not(mst_llrd_sm_dst_rdy);
  IP2Bus_MstRd_dst_dsc_n <= '1' ; --  do not throttle data

  -- implement a simple state machine to enable the
  -- read locallink interface to transfer data
  LLINK_RD_SM_PROCESS : process( Bus2IP_Clk ) is
  begin

    if ( Bus2IP_Clk'event and Bus2IP_Clk = '1' ) then
      if ( Bus2IP_Reset = '1' ) then

        -- reset condition
        mst_llrd_sm_state   <= LLRD_IDLE;
        mst_llrd_sm_dst_rdy <= '0';

      else

        -- default condition
        mst_llrd_sm_state   <= LLRD_IDLE;
        mst_llrd_sm_dst_rdy <= '0';

        -- state transition
        case mst_llrd_sm_state is

          when LLRD_IDLE =>
            if ( mst_cmd_sm_start_rd_llink = '1') then
              mst_llrd_sm_state <= LLRD_GO;
            else
              mst_llrd_sm_state <= LLRD_IDLE;
            end if;

          when LLRD_GO =>
            -- done, end of packet
            if ( mst_llrd_sm_dst_rdy    = '1' and
                 Bus2IP_MstRd_src_rdy_n = '0' and
                 Bus2IP_MstRd_eof_n     = '0' ) then
              mst_llrd_sm_state   <= LLRD_IDLE;
            -- not done yet, continue receiving data
            else
              mst_llrd_sm_state   <= LLRD_GO;
              mst_llrd_sm_dst_rdy <= '1';
            end if;

          when others =>
            mst_llrd_sm_state <= LLRD_IDLE;

        end case;

      end if;
    else
      null;
    end if;

  end process LLINK_RD_SM_PROCESS;

  -- user logic master write locallink interface assignments
  IP2Bus_MstWr_src_rdy_n <= not(mst_llwr_sm_src_rdy);
  IP2Bus_MstWr_src_dsc_n <= '1'; -- do not throttle data
  IP2Bus_MstWr_rem       <= (others => '0');
  IP2Bus_MstWr_sof_n     <= not(mst_llwr_sm_sof);
  IP2Bus_MstWr_eof_n     <= not(mst_llwr_sm_eof);

  -- implement a simple state machine to enable the
  -- write locallink interface to transfer data
  LLINK_WR_SM_PROC : process( Bus2IP_Clk ) is
    constant BYTES_PER_BEAT : integer := C_MST_DWIDTH/8;
  begin

    if ( Bus2IP_Clk'event and Bus2IP_Clk = '1' ) then
      if ( Bus2IP_Reset = '1' ) then

        -- reset condition
        mst_llwr_sm_state   <= LLWR_IDLE;
        mst_llwr_sm_src_rdy <= '0';
        mst_llwr_sm_sof     <= '0';
        mst_llwr_sm_eof     <= '0';
        mst_llwr_byte_cnt   <= 0;

      else

        -- default condition
        mst_llwr_sm_state   <= LLWR_IDLE;
        mst_llwr_sm_src_rdy <= '0';
        mst_llwr_sm_sof     <= '0';
        mst_llwr_sm_eof     <= '0';
        mst_llwr_byte_cnt   <= 0;

        -- state transition
        case mst_llwr_sm_state is

          when LLWR_IDLE =>
            if ( mst_cmd_sm_start_wr_llink = '1' and mst_cntl_burst = '0' ) then
              mst_llwr_sm_state <= LLWR_SNGL_INIT;
            elsif ( mst_cmd_sm_start_wr_llink = '1' and mst_cntl_burst = '1' ) then
              mst_llwr_sm_state <= LLWR_BRST_INIT;
            else
              mst_llwr_sm_state <= LLWR_IDLE;
            end if;

          when LLWR_SNGL_INIT =>
            mst_llwr_sm_state   <= LLWR_SNGL;
            mst_llwr_sm_src_rdy <= '1';
            mst_llwr_sm_sof     <= '1';
            mst_llwr_sm_eof     <= '1';

          when LLWR_SNGL =>
            -- destination discontinue write
            if ( Bus2IP_MstWr_dst_dsc_n = '0' and Bus2IP_MstWr_dst_rdy_n = '0' ) then
              mst_llwr_sm_state   <= LLWR_IDLE;
              mst_llwr_sm_src_rdy <= '0';
              mst_llwr_sm_eof     <= '0';
            -- single data beat transfer complete
            elsif ( mst_fifo_valid_read_xfer = '1' ) then
              mst_llwr_sm_state   <= LLWR_IDLE;
              mst_llwr_sm_src_rdy <= '0';
              mst_llwr_sm_sof     <= '0';
              mst_llwr_sm_eof     <= '0';
            -- wait on destination
            else
              mst_llwr_sm_state   <= LLWR_SNGL;
              mst_llwr_sm_src_rdy <= '1';
              mst_llwr_sm_sof     <= '1';
              mst_llwr_sm_eof     <= '1';
            end if;

          when LLWR_BRST_INIT =>
            mst_llwr_sm_state   <= LLWR_BRST;
            mst_llwr_sm_src_rdy <= '1';
            mst_llwr_sm_sof     <= '1';
            mst_llwr_byte_cnt   <= CONV_INTEGER(mst_xfer_length);

          when LLWR_BRST =>
            if ( mst_fifo_valid_read_xfer = '1' ) then
              mst_llwr_sm_sof <= '0';
            else
              mst_llwr_sm_sof <= mst_llwr_sm_sof;
            end if;
            -- destination discontinue write
            if ( Bus2IP_MstWr_dst_dsc_n = '0' and
                 Bus2IP_MstWr_dst_rdy_n = '0' ) then
              mst_llwr_sm_state   <= LLWR_IDLE;
              mst_llwr_sm_src_rdy <= '1';
              mst_llwr_sm_eof     <= '1';
            -- last data beat write
            elsif ( mst_fifo_valid_read_xfer = '1' and
                   (mst_llwr_byte_cnt-BYTES_PER_BEAT) <= BYTES_PER_BEAT ) then
              mst_llwr_sm_state   <= LLWR_BRST_LAST_BEAT;
              mst_llwr_sm_src_rdy <= '1';
              mst_llwr_sm_eof     <= '1';
            -- wait on destination
            else
              mst_llwr_sm_state   <= LLWR_BRST;
              mst_llwr_sm_src_rdy <= '1';
              -- decrement write transfer counter if it's a valid write
              if ( mst_fifo_valid_read_xfer = '1' ) then
                mst_llwr_byte_cnt <= mst_llwr_byte_cnt - BYTES_PER_BEAT;
              else
                mst_llwr_byte_cnt <= mst_llwr_byte_cnt;
              end if;
            end if;

          when LLWR_BRST_LAST_BEAT =>
            -- destination discontinue write
            if ( Bus2IP_MstWr_dst_dsc_n = '0' and
                 Bus2IP_MstWr_dst_rdy_n = '0' ) then
              mst_llwr_sm_state   <= LLWR_IDLE;
              mst_llwr_sm_src_rdy <= '0';
            -- last data beat done
            elsif ( mst_fifo_valid_read_xfer = '1' ) then
              mst_llwr_sm_state   <= LLWR_IDLE;
              mst_llwr_sm_src_rdy <= '0';
            -- wait on destination
            else
              mst_llwr_sm_state   <= LLWR_BRST_LAST_BEAT;
              mst_llwr_sm_src_rdy <= '1';
              mst_llwr_sm_eof     <= '1';
            end if;

          when others =>
            mst_llwr_sm_state <= LLWR_IDLE;

        end case;

      end if;
    else
      null;
    end if;

  end process LLINK_WR_SM_PROC;

  -- local srl fifo for data storage
  mst_fifo_valid_write_xfer <= not(Bus2IP_MstRd_src_rdy_n) and mst_llrd_sm_dst_rdy;
  mst_fifo_valid_read_xfer  <= not(Bus2IP_MstWr_dst_rdy_n) and mst_llwr_sm_src_rdy;


  ------------------------------------------
  -- Example code to drive IP to Bus signals
  ------------------------------------------
  IP2Bus_Data  <= slv_ip2bus_data when slv_read_ack = '1' else
                  mst_ip2bus_data when mst_read_ack = '1' else
                  (others => '0');

  IP2Bus_WrAck <= slv_write_ack or mst_write_ack;
  IP2Bus_RdAck <= slv_read_ack or mst_read_ack;
  IP2Bus_Error <= '0';

  -- connect to smmu

  smmu : entity smmu_core
    generic map (
        CPU_LEFT => CPU_LEFT,
        SP_LEFT => SP_LEFT,
        DATA_LEFT => DATA_LEFT,
        REF_LEFT => LOG2_NUM_ENTRIES - 1,
        NUM_ENTRIES => NUM_ENTRIES,
        LOG2_NUM_CACHE_LINES_PER_PAGE => LOG2_NUM_CACHE_LINES_PER_PAGE,
        LOG2_NUM_CACHE_WORDS_PER_LINE => LOG2_NUM_CACHE_WORDS_PER_LINE,
        INITIATE_POINT => NUM_ENTRIES - INITIATE_BOOST,
        NUM_READ_ONLY_ENTRIES => NUM_READ_ONLY_ENTRIES,
        ACTIVE_LEFT => NUM_ACTIVE_BITS - 1 )
    port map (
        cpu_address => cpu_address,
        cpu_data_in => cpu_data_in,
        cpu_data_out => cpu_data_out,
        cpu_copy_loc => cpu_copy_loc,
        cpu_size => cpu_size,
        cpu_write => cpu_write,
        cpu_read => cpu_read,
        cpu_open => cpu_open,
        cpu_close => cpu_close,
        cpu_ack => cpu_ack,
        cpu_ben => cpu_ben,
        cpu_read_only => select_read_only,
        dbg_code => dbg_code,

        cache_active => cache_enable,
        cpu_cache_set_perm => cpu_cache_set_perm,
        cpu_cache_invalidate => cpu_cache_invalidate,

        -- Add Data LMB interface here
        -- Control lines from smmu to IPIF
        mst_cntl_rd_req => mst_cntl_rd_req,
        mst_cntl_wr_req => mst_cntl_wr_req,
        mst_cntl_bus_lock => mst_cntl_bus_lock,
        mst_cntl_burst => mst_cntl_burst,
        mst_ip2bus_addr => mst_ip2bus_addr,
        mst_ip2bus_be => mst_ip2bus_be,
        mst_xfer_length => mst_xfer_length,
        mst_go => mst_go ,
        -- Status lines from IPIF to smmu
        mst_cmd_sm_set_done => mst_cmd_sm_set_done,
        mst_cmd_sm_set_error => mst_cmd_sm_set_error,
        mst_cmd_sm_set_timeout => mst_cmd_sm_set_timeout,
        mst_cmd_sm_busy => mst_cmd_sm_busy,
        mst_cmd_sm_clr_go => mst_cmd_sm_clr_go,
        -- Data interface
        output_next => mst_fifo_valid_read_xfer,
        input_next => mst_fifo_valid_write_xfer,
        data_output => IP2Bus_MstWr_d ,
        data_input => Bus2IP_MstRd_d ,

        match_mask => USE_MASK,
        -- Clock
        clk => Bus2IP_Clk,
        reset => Bus2IP_Reset ) ;


    -- Debugging code
    process ( Bus2IP_Clk , Bus2IP_Reset ) is
    begin
        if ( Bus2IP_Reset = '1' )
        then
            global_counter <= ( others => '0' ) ;
            result_counter <= ( others => '0' ) ;
            result_timer <= ( others => '0' ) ;
            match_counter <= ( others => '0' ) ;
            miss_counter <= ( others => '0' ) ;
            ctrl_mode <= '0' ;
            cache_ctrl_mode <= '0' ;
            ctrl_reg <= "000" ;
            enable_counters <= '0' ;
            select_read_only <= '0' ;
            cache_enable <= '0';

        elsif ( Bus2IP_Clk = '1' )
        and ( Bus2IP_Clk'event )
        then
            --if ( cpu_ack = '1' )
            -- then
            d_cpu_ack <= cpu_ack ;
            d_enable_counters <= enable_counters ;
            d_is_read_write <= is_read_write ;

            if ( d_cpu_ack = '1' )
            then
                if ( d_enable_counters = '1' )
                then
                    if ( d_is_read_write = '1' )
                    then
                        if ( result_timer ( 8 downto 1 ) = x"00" )
                        then
                            -- That's a match!
                            match_counter <= match_counter + 1 ;
                        else
                            -- Not a match
                            miss_counter <= miss_counter + 1 ;
                        end if ;
                    else
                        -- Not a read or write operation
                        -- (open or close..)
                        result_counter <= result_counter + 1 ;
                    end if ;
                end if ;
                enable_counters <= '0' ;
            end if ;
            if ( d_enable_counters = '1' )
            then
                result_timer <= result_timer + 1 ;
            end if ;

            ctrl_ack <= '0' ;
            sp_ack <= '0' ;
            ctrl_sp_ack <= '0' ;

            if ( ctrl_select = '1' ) 
            then
                -- Write on control memory -
                -- program copy location and size registers
                -- and enter control mode
                -- The next read/write on lmb memory will trigger
                -- an open (if read) or close (if write)
                -- To do an open:
                --   write size, copy_loc to control memory
                --   read <wanted address>
                --   value obtained = reference
                -- To do a close
                --   write <any lmb address>, reference
                -- Read on control memory -
                --   address 0: status register
                --   address 4: global counter
                --   address 8: match counter
                --   address c: miss counter
                if ( LMB_WriteStrobe = '1' )
                then
                    case LMB_ABus_int ( 4 downto 2 ) is
                    when "000" =>
                        cpu_copy_loc <= LMB_WriteDBus ( 31 - SP_LEFT to 31 ) ;
                        cpu_size <= LMB_WriteDBus ( 15 - SP_LEFT to 15 ) ;
                        select_read_only <= LMB_WriteDBus ( 0 ) ;
                        ctrl_mode <= '1' ;
                        cache_ctrl_mode <= '0' ;
                    when "011" =>
                        if ( LOG2_NUM_CACHE_WORDS_PER_LINE /= 0 )
                        then
                            cache_enable <= LMB_WriteDBus ( 31 ) ;
                            cache_ctrl_mode <= LMB_WriteDBus ( 30 );
                            ctrl_mode <= LMB_WriteDBus ( 30 );
                        end if;
                    when others =>
                        null ;
                    end case;
                end if ;
                ctrl_reg <= LMB_ABus_int ( 4 downto 2 ) ;
                ctrl_ack <= '1' ;
                ctrl_sp_ack <= '1' ;
                enable_counters <= '0' ;
            end if;
            if (( sp_select = '1' ) 
            and ENABLE_EXTRA_SCRATCHPAD )
            then
                sp_ack <= '1' ;
                ctrl_sp_ack <= '1' ;
            end if ;
            if ( lmb_select = '1' )
            then
                -- Read/write on lmb memory -
                enable_counters <= '0' ;
                if ( LMB_ABus_int ( 31 downto NUM_ACTIVE_BITS ) = USE_MASK )
                then
                    -- always exit control mode
                    ctrl_mode <= '0' ;
                    cache_ctrl_mode <= '0' ;
                    -- count read/write operations for match/miss counter
                    is_read_write <= not ctrl_mode ;
                    -- count (and time) all types of operation
                    enable_counters <= '1' ;
                    result_timer <= ( others => '0' ) ;
                end if ;
            end if ;
            global_counter <= global_counter + 1 ;
        end if ;
    end process ;

    status_reg ( 15 downto 0 ) <= result_timer ( 15 downto 0 ) ;
    status_reg ( 31 downto 16 ) <= result_counter ( 15 downto 0 ) ;


    -- Top-level port assignments
    cpu_address <= LMB_ABus_int ;
    cpu_write <= lmb_select and LMB_WriteStrobe and not ctrl_mode ;
    cpu_read <= lmb_select
            and ( not LMB_WriteStrobe ) and not ctrl_mode ;
    cpu_open <= lmb_select and ( not LMB_WriteStrobe ) 
            and ctrl_mode and not cache_ctrl_mode ;
    cpu_close <= lmb_select and LMB_WriteStrobe 
            and ctrl_mode and not cache_ctrl_mode ;
    cpu_cache_invalidate <= lmb_select and ( not LMB_WriteStrobe ) 
            and ctrl_mode and cache_ctrl_mode ;
    cpu_cache_set_perm <= lmb_select and LMB_WriteStrobe 
            and ctrl_mode and cache_ctrl_mode ;
    cpu_data_in <= LMB_WriteDBus ;
    cpu_ben <= lmb_we;

    Sl_DBus <= cpu_data_out when ( ctrl_sp_ack = '0' )
                else sp_cpu_data_out when ( sp_ack = '1' )
                else global_counter when ( ctrl_reg = "001" )
                else match_counter when ( ctrl_reg = "010" )
                else miss_counter when ( ctrl_reg = "011" )
                else x"dbdb" & dbg_code & dbg_code when ( ctrl_reg = "100" )
                else configuration_register_2 when (( ctrl_reg = "110" )
                            and ( LOG2_NUM_CACHE_WORDS_PER_LINE /= 0 ))
                else configuration_register when ( ctrl_reg = "111" )
                else status_reg ;
   
    Sl_Ready <= cpu_ack or ctrl_sp_ack ;
    LMB_ABus_int <= LMB_ABus ;
    lmb_select <= LMB_AddrStrobe and not ( ctrl_select or sp_select ) ;
    sp_select <= LMB_AddrStrobe
        when (( LMB_ABus_int ( CPU_LEFT downto NUM_ACTIVE_BITS ) = SP_MASK )
            and ENABLE_EXTRA_SCRATCHPAD )
        else '0' ;
    ctrl_select <= LMB_AddrStrobe
        when LMB_ABus_int ( CPU_LEFT downto NUM_ACTIVE_BITS ) = CONTROL_MASK
        else '0' ;

    lmb_we(0) <= LMB_BE(0) and LMB_WriteStrobe and ( lmb_select or sp_select );
    lmb_we(1) <= LMB_BE(1) and LMB_WriteStrobe and ( lmb_select or sp_select );
    lmb_we(2) <= LMB_BE(2) and LMB_WriteStrobe and ( lmb_select or sp_select );
    lmb_we(3) <= LMB_BE(3) and LMB_WriteStrobe and ( lmb_select or sp_select );

    assert LMB_Clk = Bus2IP_Clk ;

    -- Scratchpad without table
    -- This is separate from the SMMU. It is used in the test program
    -- to store memory timings and the SMMU allocation stack. It's
    -- not necessary for anything else.
    enabled : if ( ENABLE_EXTRA_SCRATCHPAD )
    generate
        lane_gen : for lane in 0 to 3
        generate 
            lane_block : block
                constant right  : Natural := 8 * lane ;
                constant left   : Natural := right + 7 ;
                signal write    : std_logic ;
            begin
                -- Control for this byte lane
                write <= LMB_WriteStrobe and sp_select 
                        and cpu_ben ( lane ) ;

                -- Memory for this byte lane
                scratchpad : RAMB16_S9
                    port map (
                        addr => cpu_address ( SP_LEFT + 2 downto 2 ),
                        di => cpu_data_in ( left downto right ),
                        dip => "0",
                        do => sp_cpu_data_out ( left downto right ),
                        dop => open,
                        clk => Bus2IP_Clk,
                        en => '1',
                        ssr => Bus2IP_Reset,
                        we => write ) ;
            end block lane_block ;
        end generate lane_gen ;
    end generate enabled ;

    configuration_register ( 31 downto 20 ) <= x"103"; -- version 1.03
    configuration_register ( 19 downto 16 ) <= 
                conv_std_logic_vector ( LOG2_NUM_ENTRIES - 1 , 4 ) ;
    configuration_register ( 15 downto 12 ) <= 
                conv_std_logic_vector ( NUM_ACTIVE_BITS - 17, 4 ) ;
    configuration_register ( 11 downto 8 ) <= 
                conv_std_logic_vector ( NUM_READ_ONLY_ENTRIES , 4 ) ;
    configuration_register ( 7 downto 4 ) <= 
                conv_std_logic_vector ( SP_LEFT , 4 ) ;
    configuration_register ( 3 ) <= 
                '1' when ENABLE_EXTRA_SCRATCHPAD else '0';
    configuration_register ( 2 ) <= 
                '1' when ( LOG2_NUM_CACHE_WORDS_PER_LINE /= 0 ) else '0';
    configuration_register ( 1 downto 0 ) <= "00" ; -- reserved bits

    configuration_register_2 ( 31 downto 28 ) <= 
                conv_std_logic_vector ( 
                        LOG2_NUM_CACHE_LINES_PER_PAGE - 1, 4 );
    configuration_register_2 ( 27 downto 24 ) <= 
                conv_std_logic_vector ( 
                        LOG2_NUM_CACHE_WORDS_PER_LINE - 1, 4 );
    configuration_register_2 ( 19 downto 0 ) <= ( others => '0' ) ;


end IMP;
